import base64
from io import StringIO, BytesIO

from dash import Input, Output, State, no_update, callback_context, MATCH, ALL
from dash import dcc, html

from dash.exceptions import PreventUpdate

from plate_planner.plate import PlateFactory

import dash_bootstrap_components as dbc
from dash_bootstrap_templates import load_figure_template

import pandas as pd

# components ids
from app.constants.component_ids import DashIdPlateDesigner, DashIdMisc, DashIdStore, DashIdPlateLib, DashIdStudy


load_figure_template(["journal", "journal_dark"])

def remove_dynamic_row(triggered_id, row_container, remove_btn_type):
    """Removes a dynamic row from a container based on the triggered remove button's ID.

    This function iterates through a list of rows (each represented as a dictionary of components),
    identifies the row that contains a remove button with the specific index and type, and removes it from the list.

    Args:
        triggered_id (dict): The ID of the component that triggered the callback. Expected to be a dictionary with
                             keys 'type' and 'index', where 'index' corresponds to the row to be removed.
        row_container (list): The list of rows (container) from which a row is to be removed. Each row is expected
                              to be a dictionary with a specific structure.
        remove_btn_type (str): The type of the remove button to be matched in the 'id' of the component.

    Returns:
        list: A new list of rows with the specified row removed.

    Example:
        triggered_id = {"type": "custom-remove-btn-type", "index": "row-3"}
        row_container = [...]
        updated_container = remove_dynamic_row(triggered_id, row_container, "custom-remove-btn-type")
    """
    index_to_remove = triggered_id.get("index")

    print(f"INDEX TO REMOVE: {index_to_remove}")

    new_row_container = []
    for row in row_container:
        component_ids = []
        for col in row.get('props', {}).get('children', []):
            for comp in col.get('props', {}).get('children', []):
                if isinstance(comp, dict):
                    comp_id = comp.get('props', {}).get('id', {})
                    if comp_id:
                        component_ids.append(comp_id)

        remove_button_present = any(
            comp_id.get('type') == remove_btn_type and comp_id.get('index') == index_to_remove
            for comp_id in component_ids
        )

        if not remove_button_present:
            new_row_container.append(row)

    return new_row_container



def register_callbacks(app):

    # Plate page loads
    @app.callback(
            [
                Output(DashIdPlateLib.PLATE_LIB_DAG.value, "rowData"),
            ],
            [
                Input(DashIdMisc.LOCATION.value, "pathname")
            ],
            [
                State(DashIdStore.PLATE_LIBRARY.value, "data")
            ]
    )
    def page_load(url, plate_lib_store):

        if url == "/plates":
            print(plate_lib_store.keys())
             # --- Update library table ---
            # Create plate objects from store
            plates = [PlateFactory.dict_to_plate(plate_dict) for name, plate_dict in plate_lib_store.items()]
            # Create dataframe for lib table
            summaries = [plate.summary_dict() for plate in plates]
            df = pd.DataFrame(summaries)
            df.insert(0, value=list(plate_lib_store.keys()), column="plate_name")

            return [df.to_dict("records")]

        raise PreventUpdate
    
    # Add a QC sample definition
    @app.callback(
        [
            Output(DashIdPlateDesigner.QC_SAMPLE_DIV.value, "children", allow_duplicate=True)
        ],
        [
            Input(DashIdPlateDesigner.ADD_QC_SAMPLE_BTN.value, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.QC_SAMPLE_DIV.value, "children")
        ],
        prevent_initial_call='initial_duplicate'
    )
    def add_qc_sample_input(n_clicks, qc_sample_rows):

        if not callback_context.triggered_id == DashIdPlateDesigner.ADD_QC_SAMPLE_BTN.value:
            raise PreventUpdate
        
        qc_sample_row = dbc.Row(
                            [
                                dbc.Label(f"{n_clicks}", width="auto"),
                                dbc.Col(
                                    [  
                                        dbc.Input(
                                            placeholder="Name",
                                            id={
                                                "type": DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.QC_SAMPLE_INDEX.value}_{n_clicks}"
                                            },
                                            valid=False,
                                            invalid=True
                                        ),
                                    ],
                                    width=6
                                ),
                                dbc.Col(
                                    [
                                        dbc.Input(
                                            placeholder="Abbreviation",
                                            id={
                                                "type": DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.QC_SAMPLE_INDEX.value}_{n_clicks}"
                                            },
                                            valid=False,
                                            invalid=True
                                        ),
                                    ]
                                ),
                                dbc.Col(
                                    [
                                        dbc.Button(
                                            "Remove",
                                            color="primary",
                                            outline=True,
                                            id={
                                                    "type": DashIdPlateDesigner.QC_SAMPLE_REMOVE_BTN_TYPE.value,
                                                    "index": f"{DashIdPlateDesigner.QC_SAMPLE_INDEX.value}_{n_clicks}"
                                            },
                                        ),
                                    ],
                                    width="auto"
                                )
                            ],
                            className="mb-2"
                        )
        
        print(f"INDEX: {DashIdPlateDesigner.QC_SAMPLE_INDEX.value}_{n_clicks}")
        
        qc_sample_rows.append(qc_sample_row)

        return [qc_sample_rows]
    
    # Remove a QC sample defintion
    # NOTE RESET all defined QC rounds to guarantee consistency
    # TODO handle this more elegantly perhaps
    @app.callback(
        [
            Output(DashIdPlateDesigner.QC_SAMPLE_DIV.value, "children", allow_duplicate=True),
            Output(DashIdPlateDesigner.SPEC_ROUND_DIV.value, "children", allow_duplicate=True)
        ],
        [
            Input({"type": DashIdPlateDesigner.QC_SAMPLE_REMOVE_BTN_TYPE.value, "index": ALL}, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.QC_SAMPLE_DIV.value, "children")
        ],
        prevent_initial_call='initial_duplicate'
    )
    def remove_qc_sample_input(n_clicks_list, qc_sample_rows):

        remove_btn_type = DashIdPlateDesigner.QC_SAMPLE_REMOVE_BTN_TYPE.value

        # Check if any button was clicked
        if not any(n_clicks and n_clicks > 0 for n_clicks in n_clicks_list):
            raise PreventUpdate

        if not callback_context.triggered:
            raise PreventUpdate

        # Get the id of the button that triggered the callback
        triggered_id = callback_context.triggered_id


        # If the callback was triggered by a pattern-matching callback, triggered_id is a dictionary
        if not isinstance(triggered_id, dict):
            raise PreventUpdate

        new_qc_sample_rows = remove_dynamic_row(triggered_id, qc_sample_rows, remove_btn_type)

        return new_qc_sample_rows, []
    
    # Check validity of input for QC sample definitions
    @app.callback(
    [
        Output({'type': DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, 'index': MATCH}, 'valid'),
        Output({'type': DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, 'index': MATCH}, 'invalid'),
        Output({'type': DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, 'index': MATCH}, 'valid'),
        Output({'type': DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, 'index': MATCH}, 'invalid')
    ],
    [
        Input({'type': DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, 'index': ALL}, 'value'),
        Input({'type': DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, 'index': ALL}, 'value')
    ],
    [
        State({'type': DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, 'index': MATCH}, 'id'),
        State({'type': DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, 'index': MATCH}, 'id')
    ]
    )
    def update_input_validity(name_values, abbrev_values, name_id, abbrev_id):
        if not callback_context.triggered:
            raise PreventUpdate
        
        prop_id = callback_context.triggered[0]['prop_id']
        value = callback_context.triggered[0]['value']

        if 'qc-sample-name-input-type' in prop_id:
            # Logic for "Name" input field
            is_valid = bool(value)
            return is_valid, not is_valid, no_update, no_update
        elif 'qc-sample-abbrev-input-type' in prop_id:
            # Logic for "Abbreviation" input field
            is_valid = len(value) <= 3
            return no_update, no_update, is_valid, not is_valid
        else:
            raise PreventUpdate
        

    # Add spec round definition
    @app.callback(
        [
            Output(DashIdPlateDesigner.SPEC_ROUND_DIV.value, "children"),
            Output(DashIdPlateDesigner.SPEC_ROUND_ALERT.value, "is_open")
        ],
        [
            Input(DashIdPlateDesigner.ADD_SPEC_ROUND_BTN.value, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.SPEC_ROUND_DIV.value, "children"),

            State({"type": DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, "index": ALL}, "value")
        ]
    )
    def add_spec_qc_round(n_clicks, spec_round_rows, qc_sample_names, qc_sample_abbrev):

        if None in qc_sample_abbrev or len(qc_sample_abbrev) == 0:
            return no_update, True

        options = [abbrev for abbrev in qc_sample_abbrev]
        options = options if options is not None else []

        if not callback_context.triggered:
            raise PreventUpdate

        spec_round_row = dbc.Row(
                            [
                                # dbc.Label(f"{n_clicks}"),
                                dbc.Col(
                                    [
                                        dbc.InputGroup(
                                            [
                                                dbc.InputGroupText("Round"),
                                                dbc.Input(
                                                    placeholder="Name",
                                                    type="number",
                                                    min=1,
                                                    value=1,
                                                    id={
                                                        "type": DashIdPlateDesigner.SPEC_ROUND_INPUT_TYPE.value,
                                                        "index": f"{DashIdPlateDesigner.SPEC_ROUND_INDEX.value}_{n_clicks}"
                                                    },
                                                ),
                                            ]
                                        )
                                    ],
                                ),
                                dbc.Col(
                                    [
                                        dcc.Dropdown(
                                            # placeholder="Abbreviation",
                                            options=options,
                                            id={
                                                "type": DashIdPlateDesigner.SPEC_ROUND_SELECT_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.SPEC_ROUND_INDEX.value}_{n_clicks}"
                                            },

                                        ),
                                    ],
                                ),
                                dbc.Col(
                                    [
                                        dbc.Button(
                                            "Remove",
                                            color="primary",
                                            outline=True,
                                            id={
                                                "type": DashIdPlateDesigner.REMOVE_SPEC_ROUND_BTN_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.SPEC_ROUND_INDEX.value}_{n_clicks}"
                                            }
                                        ),
                                    ],
                                    width="auto"
                                )
                            ],
                            className="mb-2"
                        )
                
        spec_round_rows.append(spec_round_row)

        return spec_round_rows, False
    
    # Remove spec round definition
    @app.callback(
        [
            Output(DashIdPlateDesigner.SPEC_ROUND_DIV.value, "children", allow_duplicate=True)
        ],
        [
            Input({"type": DashIdPlateDesigner.REMOVE_SPEC_ROUND_BTN_TYPE.value, "index": ALL}, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.SPEC_ROUND_DIV.value, "children")
        ],
        prevent_initial_call='initial_duplicate'
    )
    def remove_spec_round(n_clicks_list, spec_rounds_rows):

        remove_btn_type = DashIdPlateDesigner.REMOVE_SPEC_ROUND_BTN_TYPE.value

        # Check if any button was clicked
        if not any(n_clicks and n_clicks > 0 for n_clicks in n_clicks_list):
            raise PreventUpdate

        if not callback_context.triggered:
            raise PreventUpdate

        # Get the id of the button that triggered the callback
        triggered_id = callback_context.triggered_id

        # If the callback was triggered by a pattern-matching callback, triggered_id is a dictionary
        if not isinstance(triggered_id, dict):
            raise PreventUpdate

        new_spec_rounds_rows = remove_dynamic_row(triggered_id, spec_rounds_rows, remove_btn_type)

        return [new_spec_rounds_rows]
    
    # Add start round definitions
    @app.callback(
        [
            Output(DashIdPlateDesigner.START_ROUND_DIV.value, "children"),
            Output(DashIdPlateDesigner.START_ROUND_ALERT.value, "is_open")
        ],
        [
            Input(DashIdPlateDesigner.ADD_START_ROUND_BTN.value, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.START_ROUND_DIV.value, "children"),

            State({"type": DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, "index": ALL}, "value"),

            State(DashIdStore.CURRENT_PLATE_DESIGN.value, "data")
        ]
    )
    def add_start_qc_round(n_clicks, start_round_rows, qc_sample_names, qc_sample_abbrev, plate_json):

        if None in qc_sample_abbrev or len(qc_sample_abbrev) == 0:
            return no_update, True

        options = [abbrev for abbrev in qc_sample_abbrev]
        options = options if options is not None else []

        if not callback_context.triggered:
            raise PreventUpdate
        
        current_plate_design = PlateFactory.dict_to_plate(plate_json)

        default_position = len(start_round_rows)  + 1

        position_options = []
        for label, value in  zip(current_plate_design._alphanumerical_coordinates, range(1, current_plate_design.capacity+1)):
            position_options.append({"label": label, "value": value})

        start_round_row = dbc.Row(
                            [
                                # dbc.Label(f"{len(spec_round_rows)+1}", width="auto"),
                                dbc.Col(
                                    [
                                        dbc.InputGroup(
                                            [
                                                dbc.InputGroupText("Position"),
                                                dbc.Select(
                                                    id={
                                                        "type": DashIdPlateDesigner.START_ROUND_POS_SELECT_TYPE.value,
                                                        "index": f"{DashIdPlateDesigner.START_ROUND_INDEX.value}_{n_clicks}"
                                                    },
                                                    options=position_options,
                                                    value=default_position
                                                )
                                            ]
                                        )
                                    ],
                                ),
                                dbc.Col(
                                    [
                                        dcc.Dropdown(
                                            # placeholder="Abbreviation",
                                            options=options,
                                            id={
                                                "type": DashIdPlateDesigner.START_ROUND_SELECT_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.START_ROUND_INDEX.value}_{n_clicks}"
                                            },

                                        ),
                                    ],
                                ),
                                dbc.Col(
                                    [
                                        dbc.Button(
                                            "Remove",
                                            color="primary",
                                            outline=True,
                                            id={
                                                "type": DashIdPlateDesigner.REMOVE_START_ROUND_BTN_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.START_ROUND_INDEX.value}_{n_clicks}"
                                            }
                                        ),
                                    ],
                                    width="auto"
                                )
                            ],
                            className="mb-2"
                        )
        
        start_round_rows.append(start_round_row)

        return start_round_rows, False
    
    # Remove start round definition
    @app.callback(
        [
            Output(DashIdPlateDesigner.START_ROUND_DIV.value, "children", allow_duplicate=True)
        ],
        [
            Input({"type": DashIdPlateDesigner.REMOVE_START_ROUND_BTN_TYPE.value, "index": ALL}, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.START_ROUND_DIV.value, "children")
        ],
        prevent_initial_call='initial_duplicate'
    )
    def remove_start_round(n_clicks_list, spec_rounds_rows):

        remove_btn_type = DashIdPlateDesigner.REMOVE_START_ROUND_BTN_TYPE.value

        # Check if any button was clicked
        if not any(n_clicks and n_clicks > 0 for n_clicks in n_clicks_list):
            raise PreventUpdate

        if not callback_context.triggered:
            raise PreventUpdate

        # Get the id of the button that triggered the callback
        triggered_id = callback_context.triggered_id

        # If the callback was triggered by a pattern-matching callback, triggered_id is a dictionary
        if not isinstance(triggered_id, dict):
            raise PreventUpdate

        new_spec_rounds_rows = remove_dynamic_row(triggered_id, spec_rounds_rows, remove_btn_type)

        return [new_spec_rounds_rows]
    
    # Add end round definitions
    @app.callback(
        [
            Output(DashIdPlateDesigner.END_ROUND_DIV.value, "children"),
            Output(DashIdPlateDesigner.END_ROUND_ALERT.value, "is_open")
        ],
        [
            Input(DashIdPlateDesigner.ADD_END_ROUND_BTN.value, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.END_ROUND_DIV.value, "children"),

            State({"type": DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, "index": ALL}, "value"),

            State(DashIdStore.CURRENT_PLATE_DESIGN.value, "data")
        ]
    )
    def add_end_qc_round(n_clicks, end_round_rows, qc_sample_names, qc_sample_abbrev, plate_json):


        if None in qc_sample_abbrev or len(qc_sample_abbrev) == 0:
            return no_update, True

        options = [abbrev for abbrev in qc_sample_abbrev]
        options = options if options is not None else []

        if not callback_context.triggered:
            raise PreventUpdate
        
        current_plate_design = PlateFactory.dict_to_plate(plate_json)

        position_options = []
        for label, value in  zip(current_plate_design._alphanumerical_coordinates[::-1], range(1, current_plate_design.capacity+1)):
            position_options.append({"label": label, "value": value})

        default_position = len(end_round_rows) + 1
        
        end_round_row = dbc.Row(
                            [
                                # dbc.Label(f"{len(spec_round_rows)+1}", width="auto"),
                                dbc.Col(
                                    [
                                        dbc.InputGroup(
                                            [
                                                dbc.InputGroupText("Position"),
                                                dbc.Select(
                                                    id={
                                                        "type": DashIdPlateDesigner.END_ROUND_POS_SELECT_TYPE.value,
                                                        "index": f"{DashIdPlateDesigner.END_ROUND_INDEX.value}_{n_clicks}"
                                                    },
                                                    options=position_options,
                                                    value=default_position
                                                )
                                            ]
                                        )
                                    ],
                                ),
                                dbc.Col(
                                    [
                                        dcc.Dropdown(
                                            # placeholder="Abbreviation",
                                            options=options,
                                            id={
                                                "type": DashIdPlateDesigner.END_ROUND_SELECT_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.END_ROUND_INDEX.value}_{n_clicks}"
                                            },

                                        ),
                                    ],
                                ),
                                dbc.Col(
                                    [
                                        dbc.Button(
                                            "Remove",
                                            color="primary",
                                            outline=True,
                                            id={
                                                "type": DashIdPlateDesigner.REMOVE_END_ROUND_BTN_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.END_ROUND_INDEX.value}_{n_clicks}"
                                            }
                                        ),
                                    ],
                                    width="auto"
                                )
                            ],
                            className="mb-2"
                        )
        
        end_round_rows.append(end_round_row)

        return end_round_rows, False
    
    # Remove end round definition
    @app.callback(
        [
            Output(DashIdPlateDesigner.END_ROUND_DIV.value, "children", allow_duplicate=True)
        ],
        [
            Input({"type": DashIdPlateDesigner.REMOVE_END_ROUND_BTN_TYPE.value, "index": ALL}, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.END_ROUND_DIV.value, "children")
        ],
        prevent_initial_call='initial_duplicate'
    )
    def remove_end_round(n_clicks_list, spec_rounds_rows):

        remove_btn_type = DashIdPlateDesigner.REMOVE_END_ROUND_BTN_TYPE.value

        # Check if any button was clicked
        if not any(n_clicks and n_clicks > 0 for n_clicks in n_clicks_list):
            raise PreventUpdate

        if not callback_context.triggered:
            raise PreventUpdate

        # Get the id of the button that triggered the callback
        triggered_id = callback_context.triggered_id

        # If the callback was triggered by a pattern-matching callback, triggered_id is a dictionary
        if not isinstance(triggered_id, dict):
            raise PreventUpdate

        new_spec_rounds_rows = remove_dynamic_row(triggered_id, spec_rounds_rows, remove_btn_type)

        return [new_spec_rounds_rows]

    # Add alternating round definition
    @app.callback(
        [
            Output(DashIdPlateDesigner.ALTERNATE_ROUND_DIV.value, "children"),
            Output(DashIdPlateDesigner.ALTERNATE_ROUND_ALERT.value, "is_open")
        ],
        [
            Input(DashIdPlateDesigner.ADD_ALTERNATE_ROUND_BTN.value, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.ALTERNATE_ROUND_DIV.value, "children"),

            State({"type": DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, "index": ALL}, "value"),

            State(DashIdStore.CURRENT_PLATE_DESIGN.value, "data")
        ]
    )
    def add_alternating_qc_round(n_clicks, end_round_rows, qc_sample_names, qc_sample_abbrev, plate_json):

        if None in qc_sample_abbrev or len(qc_sample_abbrev) == 0:
            return no_update, True

        options = [abbrev for abbrev in qc_sample_abbrev]
        options = options if options is not None else []

        if not callback_context.triggered:
            raise PreventUpdate
        
        default_position = len(end_round_rows) + 1
        
        end_round_row = dbc.Row(
                            [
                                # dbc.Label(f"{len(spec_round_rows)+1}", width="auto"),
                                dbc.Col(
                                    [
                                        dbc.InputGroup(
                                            [
                                                dbc.InputGroupText("Round"),
                                                dbc.Input(
                                                    type="number",
                                                    min=1,
                                                    value=1,
                                                    id={
                                                        "type": DashIdPlateDesigner.ALTERNATE_ROUND_INPUT_TYPE.value,
                                                        "index": f"{DashIdPlateDesigner.ALTERNATE_ROUND_INDEX.value}_{n_clicks}"
                                                    },
                                                )
                                            ]
                                        )
                                    ],
                                ),
                                dbc.Col(
                                    [
                                        dcc.Dropdown(
                                            # placeholder="Abbreviation",
                                            options=options,
                                            id={
                                                "type": DashIdPlateDesigner.ALTERNATE_ROUND_SELECT_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.ALTERNATE_ROUND_INDEX.value}_{n_clicks}"
                                            },

                                        ),
                                    ],
                                ),
                                dbc.Col(
                                    [
                                        dbc.Button(
                                            "Remove",
                                            color="primary",
                                            outline=True,
                                            id={
                                                "type": DashIdPlateDesigner.REMOVE_ALTERNATE_ROUND_BTN_TYPE.value,
                                                "index": f"{DashIdPlateDesigner.ALTERNATE_ROUND_INDEX.value}_{n_clicks}"
                                            }
                                        ),
                                    ],
                                    width="auto"
                                )
                            ],
                            className="mb-2"
                        )
        
        end_round_rows.append(end_round_row)

        return end_round_rows, False
    
    # Remove alternating definition
    @app.callback(
        [
            Output(DashIdPlateDesigner.ALTERNATE_ROUND_DIV.value, "children", allow_duplicate=True)
        ],
        [
            Input({"type": DashIdPlateDesigner.REMOVE_ALTERNATE_ROUND_BTN_TYPE.value, "index": ALL}, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.ALTERNATE_ROUND_DIV.value, "children")
        ],
        prevent_initial_call='initial_duplicate'
    )
    def remove_alternating_round(n_clicks_list, spec_rounds_rows):

        remove_btn_type = DashIdPlateDesigner.REMOVE_ALTERNATE_ROUND_BTN_TYPE.value

        # Check if any button was clicked
        if not any(n_clicks and n_clicks > 0 for n_clicks in n_clicks_list):
            raise PreventUpdate

        if not callback_context.triggered:
            raise PreventUpdate

        # Get the id of the button that triggered the callback
        triggered_id = callback_context.triggered_id

        # If the callback was triggered by a pattern-matching callback, triggered_id is a dictionary
        if not isinstance(triggered_id, dict):
            raise PreventUpdate

        new_spec_rounds_rows = remove_dynamic_row(triggered_id, spec_rounds_rows, remove_btn_type)

        return [new_spec_rounds_rows]


    # Plate viz options changed
    @app.callback(
        [
            Output(DashIdPlateDesigner.PREVIEW_GRAPH.value, "figure", allow_duplicate=True)
        ],
        [
            Input(DashIdPlateDesigner.COLOR_SELECT.value, "value"),
            Input(DashIdPlateDesigner.LABEL_SELECT.value, "value"),
        ],
        prevent_initial_call="initial_duplicate"
    )
    def update_plate_viz(color_option, label_option):
        raise PreventUpdate
    
    # ##################################################################################################
    # CREATE/PREVIEW clicked
    @app.callback(
        [
            Output(DashIdPlateDesigner.PREVIEW_GRAPH.value, "figure", allow_duplicate=True),
            Output(DashIdPlateDesigner.PREVIEW_DAG.value, "rowData", allow_duplicate=True),

            Output(DashIdPlateDesigner.SIZE_SELECT.value, "value"),
            Output(DashIdPlateDesigner.SIZE_SELECT.value, "options"),

            Output(DashIdPlateDesigner.N_ROWS_SELECT.value, "value"),
            Output(DashIdPlateDesigner.N_COLUMNS_SELECT.value, "value"),

            Output(DashIdStore.PLATE_LIBRARY.value, "data"),

            Output(DashIdPlateLib.PLATE_LIB_DAG.value, "rowData", allow_duplicate=True),

            Output(DashIdStore.CURRENT_PLATE_DESIGN.value, "data"),
        ],
        [
            # Preview button
            Input(DashIdPlateDesigner.PREVIEW_BTN.value, "n_clicks"),

            # Create button
            Input(DashIdPlateDesigner.CREATE_BTN.value, "n_clicks"),

            Input(DashIdPlateDesigner.SIZE_SELECT.value, "value"),
            Input(DashIdPlateDesigner.N_ROWS_SELECT.value, "value"),
            Input(DashIdPlateDesigner.N_COLUMNS_SELECT.value, "value"),

            Input(DashIdMisc.DARK_MODE_SWITCH.value, "value"),

            # Plate viz settings
            Input(DashIdPlateDesigner.COLOR_SELECT.value, "value"),
            Input(DashIdPlateDesigner.LABEL_SELECT.value, "value"),
        ],
        [
            # QC frequency
            State(DashIdPlateDesigner.QC_SAMPLE_SPACING_INPUT.value, "value"),
            State(DashIdPlateDesigner.START_WITH_QC_CHECKBOX.value, "value"),

            # Plate size select options
            State(DashIdPlateDesigner.SIZE_SELECT.value, "options"),

            # QC sample names
            State({"type": DashIdPlateDesigner.QC_SAMPLE_NAME_INPUT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.QC_SAMPLE_ABBREV_INPUT_TYPE.value, "index": ALL}, "value"),

            # Start pattern
            State({"type": DashIdPlateDesigner.START_ROUND_POS_SELECT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.START_ROUND_SELECT_TYPE.value, "index": ALL}, "value"),

            # End
            State({"type": DashIdPlateDesigner.END_ROUND_POS_SELECT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.END_ROUND_SELECT_TYPE.value, "index": ALL}, "value"),

            # Specific
            State({"type": DashIdPlateDesigner.SPEC_ROUND_INPUT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.SPEC_ROUND_SELECT_TYPE.value, "index": ALL}, "value"),

            # Alternate
            State({"type": DashIdPlateDesigner.ALTERNATE_ROUND_INPUT_TYPE.value, "index": ALL}, "value"),
            State({"type": DashIdPlateDesigner.ALTERNATE_ROUND_SELECT_TYPE.value, "index": ALL}, "value"),

            # Store
            State(DashIdStore.PLATE_LIBRARY.value, "data"),

            # Plate name
            State(DashIdPlateDesigner.NAME_INPUT.value, "value"),

            # Url
            State(DashIdMisc.LOCATION.value, "pathname")
        ],
        prevent_initial_call="initial_duplicate"
    )
    def create_preview_plate(n_clicks_preview, n_clicks_create, plate_size, n_rows, n_cols, dark_mode_off, 
                     color_option, label_option,
                     qc_sample_spacing, start_with_qc,
                     plate_size_options,
                     qc_sample_names, qc_sample_abbrev,
                     start_positions, start_samples,
                     end_positions, end_samples,
                     spec_round_number, spec_round_samples,
                     alt_round_number, alt_round_samples,
                     plate_lib_store,
                     plate_name,
                     url
                     ):
        
        def group_values(group_index, group_samples):
            grouped_values = {}
            for index, value in zip(group_index, group_samples):
                if index in grouped_values:
                    grouped_values[index].append(value)
                else:
                    grouped_values[index] = [value]

            # Group values based on the group_indices
            for index, value in zip(spec_round_number, spec_round_samples):
                if index in grouped_values:
                    grouped_values[index].append(value)
                else:
                    grouped_values[index] = [value]

            return grouped_values
        
        if (url != "/plates"):
            raise PreventUpdate

        template = "journal" if dark_mode_off else "journal_dark"

        if callback_context.triggered_id == DashIdPlateDesigner.SIZE_SELECT.value:
            match plate_size:
                case "6":
                    n_rows = 2
                    n_cols = 3
                case "12":
                    n_rows = 3
                    n_cols = 4
                case "24":
                    n_rows = 4
                    n_cols = 6
                case "48":
                    n_rows = 6
                    n_cols = 8
                case "96":
                    n_rows = 8
                    n_cols = 12
                case "384":
                    n_rows = 16
                    n_cols = 24
                case _:
                    raise PreventUpdate
                
        if (callback_context.triggered_id == DashIdPlateDesigner.N_ROWS_SELECT.value) or (callback_context.triggered_id == DashIdPlateDesigner.N_COLUMNS_SELECT.value):

            if n_rows/2 * 3 == n_cols:
                plate_size = f"{n_rows*n_cols}"
            else:
                plate_size = f"{n_rows*n_cols} (custom)"
        
        plate_size_options[-1] = plate_size

        # if not n_clicks:
        #     raise PreventUpdate
        
        # QC SAMPLE ABBREVIATION-NAME MAPPING
        qc_config_samplenames = {abbrev: name for abbrev,name in zip(qc_sample_abbrev, qc_sample_names)}

        # SETUP QC PATTERNS
        qc_patterns = {}
    
        # START
        qc_pattern_start = [start_samples[int(index)-1] for index in start_positions]
        qc_patterns["start"] = qc_pattern_start

        # END
        qc_pattern_end = [end_samples[int(index)-1] for index in end_positions[::-1]]
        qc_patterns["end"] = qc_pattern_end

        # SPECIFIC
        grouped_values = group_values(spec_round_number, spec_round_samples)

        for key in grouped_values.keys():
            round_name = f"round_{key}"
            qc_patterns[round_name] = grouped_values[key]
            # print(f"{round_name}: {grouped_values[key]}")

        # ALTERNATING
        grouped_values = group_values(alt_round_number, alt_round_samples)
        alternating_pattern = []
        for key in grouped_values.keys():
            alternating_pattern.append(grouped_values[key])

        if alternating_pattern:
            qc_patterns["alternating"] = alternating_pattern

        # REPEAT
        

        # qc_pattern_specific =
        print(qc_patterns)

        QC_config_dict =  {
            'QC': {
                'start_with_QC_round': start_with_qc,
                'run_QC_after_n_specimens': qc_sample_spacing,
                'names': qc_config_samplenames,
                'patterns': qc_patterns
            }
        }

        if qc_patterns:
            plate = PlateFactory.create_plate(plate_dim=[n_rows, n_cols], QC_config = QC_config_dict)
        else:
            plate = PlateFactory.create_plate(plate_dim=[n_rows, n_cols],)
        
        fig = plate.as_plotly_figure(
            annotation_metadata_key=label_option,
            color_metadata_key=color_option,
            theme=template,
            show_grid=False,
            fig_height=500,
            fig_width=750,
            dark_mode=not dark_mode_off,
            colormap_discrete="Set2"
        )

        plate_df_dict = plate.as_dataframe().to_dict("records")

        if callback_context.triggered_id == DashIdPlateDesigner.CREATE_BTN.value:
            if not plate_lib_store:
                plate_lib_store = {}
            
            # --- Add plate to store ---
            plate_lib_store[plate_name] = plate.as_dict()

            # -- Update plate select options
            plate_select_options = [name for name in plate_lib_store.keys()]

            # --- Update library table ---
            # Create plate objects from store
            plates = [PlateFactory.dict_to_plate(plate_dict) for name, plate_dict in plate_lib_store.items()]
            # Create dataframe for lib table
            summaries = [plate.summary_dict() for plate in plates]
            df = pd.DataFrame(summaries)
            df.insert(0, value=list(plate_lib_store.keys()), column="plate_name")

            return fig, plate_df_dict, plate_size, plate_size_options, n_rows, n_cols, plate_lib_store, df.to_dict("records"), plate.as_dict()

        return fig, plate_df_dict, plate_size, plate_size_options, n_rows, n_cols, no_update, no_update, plate.as_dict()

    # Create plate -> Show info alert
    @app.callback(
        [
            Output(DashIdPlateDesigner.ADDED_TO_LIB_ALERT.value, "is_open"),
            Output(DashIdPlateDesigner.ADDED_TO_LIB_ALERT.value, "children"),
        ],
        [
            Input(DashIdPlateDesigner.CREATE_BTN.value, "n_clicks")
        ],
        [
            State(DashIdPlateDesigner.NAME_INPUT.value, "value"),
        ]
    )
    def add_plate_to_lib(n_clicks, plate_name):

        if n_clicks:
            emphasized_plate_name = html.Span(plate_name, style={'font-weight': 'bold'})
            message = [emphasized_plate_name, ' added to library.']

            return True, message

        raise PreventUpdate
    
    # Remove plate from library
    @app.callback(
        [
            Output(DashIdStore.PLATE_LIBRARY.value, "data", allow_duplicate=True),
            Output(DashIdPlateLib.PLATE_LIB_DAG.value, "rowData", allow_duplicate=True),
        ],
        [
            Input(DashIdPlateLib.REMOVE_PLATE_BTN.value, "n_clicks")
        ],
        [
            State(DashIdStore.PLATE_LIBRARY.value, "data"),
            State(DashIdPlateLib.PLATE_LIB_DAG.value, "selectedRows"),
            State(DashIdPlateLib.PLATE_LIB_DAG.value, "rowData"),
        ],
        prevent_initial_call="initial_duplicate"
    )
    def remove_plate_lib(n_clicks, plate_lib_store, selected, row_data):
        if not n_clicks or not selected:
            raise PreventUpdate

        try:
            # Remove the selected plate from the store
            selected_plate_name = selected[0]["plate_name"]
            if selected_plate_name in plate_lib_store:
                del plate_lib_store[selected_plate_name]

            # Update rowData for AG Grid, removing the selected row
            updated_row_data = [row for row in row_data if row['plate_name'] != selected_plate_name]

            return plate_lib_store, updated_row_data

        except Exception as e:
            print(e)
            # Log the error or handle it as necessary
            print(f"Error removing plate: {e}")
            raise PreventUpdate
        




from dash import dcc, html
import dash_bootstrap_components as dbc

from app.constants.component_ids import DashIdMisc

# color switch
dark_mode_switch = html.Span(
    [
        dbc.Label(
            className="fa fa-moon",
            # style={"color": "white"}
            # html_for="color-mode-switch"
        ),
        dbc.Switch(
            id=DashIdMisc.DARK_MODE_SWITCH.value,
            value=True,
            className="d-inline-block",
            style={"margin-left": "8px"},
            persistence=True,
            
        ),
        dbc.Label(
            className="fa fa-sun",
            # style={"color": "white"}
            # html_for="color-mode-switch"
        ),
    ],
)

navbar = dbc.NavbarSimple(
            [
                dbc.NavItem(dbc.NavLink("Home", href="/"), style={"font-size":"1.1rem"}),
                dbc.NavItem(dbc.NavLink("Plates", href="/plates"), style={"font-size":"1.1rem"}),
                dbc.NavItem(dbc.NavLink("Project", href="/project"), style={"font-size":"1.1rem"}),
                dbc.NavItem(dbc.NavLink("  ")),
                dbc.NavItem(
                    html.Div(dark_mode_switch,style={"padding-top": "10px"}),
                    
                ),
                dbc.NavItem(dbc.NavLink("  ")),
                dbc.NavItem(dbc.NavLink(
                    html.I(className="fa-brands fa-github", style={"font-size":"2.0rem"}),
                    href="https://github.com/ssi-dk/CD-MRG-PlatePlanner",
                    target="_blank"),
                ),
            ],
            brand=html.Img(src="./assets/logo.png", width="300"),
            brand_href="/",
            color="light",
            id=DashIdMisc.NAVBAR.value
        )


import dash_bootstrap_components as dbc
from dash import html

import dash

dash.register_page(
    __name__,
    path="/",
    title='PlatePlanner',
    name='Home')

footer = dbc.Row(
    dbc.Col(
        [
            html.P("PlatePlanner - Streamline your plate designs for LC-MS analysis and beyond.", className="footer-title"),
            html.A("GitHub Repository", href="https://github.com/YourRepo/PlatePlanner", className="footer-link"),
            html.A("PlatePlanner Python Package", href="https://pypi.org/project/plateplanner/", className="footer-link"),
            html.P("Developed by the PlatePlanner Team", className="footer-text")
        ],
        md=12,
        className="ms-4"
    ),
    className="mt-5 py-3 bg-light",  # Add your background color and padding here.
    # style={"width": "100%", "margin-top": "auto"}   # Ensure the footer spans the full width.
)



body = dbc.Container(
    [
        dbc.Row(
            [
                dbc.Col(
                    [
                        # html.H4("Let's plate!"),
                        html.P(
                            [
                            """
                            Welcome to PlatePlanner, a tool for designing multi-well plate layouts
                            for LC-MS analysis and beyond!
                            """,
                            ],
                            style={"font-size":"1.2rem"}
                        ),
                        # html.P(
                        #     """
                        #     PlatePlanner streamlines the process of creating custom 
                        #     plate templates, distributing samples, and generating detailed run lists and visualizations.
                            
                        #     """
                        # )
                    ],
                    width=6
                ),
               
            ],
            class_name="mt-4"
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        html.Div( style={"margin-top": "50px"}),
                    ]
                )
            ],
            class_name="mt-4"
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        html.H4("Getting started"),
                        html.Div(className="mt-4"),

                        dbc.Stack(
                            [
                                html.A(
                                    [
                                        html.I(className="fa-solid fa-pen-to-square", style={"font-size": "2.0rem"}),
                                        html.H5("Design plate templates", className="d-inline ms-2")  # Adjusted for inline display and spacing
                                    ],
                                    href="/plates",  
                                    className="text-decoration-none",  
                                )
                            ],
                            direction="horizontal",
                            gap=3
                        ),

                        

                        html.Div(
                            [
                                """
                                Create a plate template: set plate size and add desired QC samples and frequency patterns.
                                """
                            ],
                            className="mt-2"
                        ),

                        html.Div( style={"margin-top": "50px"}),

                        dbc.Stack(
                            [
                                html.A(
                                    [
                                        html.I(className="fa-solid fa-layer-group", style={"font-size": "2.0rem"}),
                                        html.H5("Distribute samples to plates", className="d-inline ms-2")  # Adjusted for inline display and spacing
                                    ],
                                    href="/project",  
                                    className="text-decoration-none",  
                                )
                            ],
                            direction="horizontal",
                            gap=3
                        ),

                        html.Div(
                            [
                                """
                                Upload a sample list and let PlatePlanner distribute the samples to your plate templates. Samples or sample groups (blocks) can be uniformly randomized across plates. 
                                """

                            ],
                            className="mt-2"
                        ),

                        html.Div( style={"margin-top": "50px"}),

                        dbc.Stack(
                            [
                                html.A(
                                    [
                                        html.I(className="fa-solid fa-file-arrow-down", style={"font-size": "2.0rem"}),
                                        html.H5("Inspect/download", className="d-inline ms-2")  # Adjusted for inline display and spacing
                                    ],
                                    href="/project",  
                                    className="text-decoration-none",  
                                )
                            ],
                            direction="horizontal",
                            gap=3
                        ),

                        html.Div(
                            [
                                """
                                Explore your new project plates - download run lists or custom plate visualization`                                """

                            ],
                            className="mt-2"
                        ),
                        
                    ],
                    width=6
                ),
            ]
        ),
        html.Div(className="mt-4 mb-4"),
    ],
)

layout = html.Div(
    [
        body,
        # footer
    ],
    # style={'display': 'flex', 'flex-direction': 'column', 'height': '100vh'}
)


from enum import Enum

class DashIdStore(Enum):
    CURRENT_PLATE_DESIGN = "current-plate-design-store"
    PLATE_LIBRARY = "plate-library-store"
    SAMPLE_LIST = "sample-list-store"
    PLATE_LAYOUTS = "plate-layout-store"
    STUDY = "study-store"

class DashIdMisc(Enum):
    DARK_MODE_SWITCH = "dark-mode-switch"
    NAVBAR = "navbar-simple"
    LOCATION = "current-page"

class DashIdPlateDesigner(Enum):
    PREVIEW_GRAPH = "plate-preview-graph"

    PREVIEW_DAG = "plate-preview-dag"

    NAME_INPUT = "plate-name-input"
    SIZE_SELECT = "plate-size-select"
    N_ROWS_SELECT = "plate-n-row-select"
    N_COLUMNS_SELECT = "plate-n-columns-select"

    # Plate preview
    COLOR_SELECT = "color-wells-select"
    LABEL_SELECT = "label-wells-select"

    CREATE_BTN = "create-plate-btn"
    PREVIEW_BTN = "preview-plate-btn"
    RESET_BTN = "reset-plate-btn"

    ## QC sample definitions
    QC_SAMPLE_SPACING_INPUT = "number-of-samples-between-qc-rounds-input"
    START_WITH_QC_CHECKBOX = "start-with-qc-checkbox"

    # container
    QC_SAMPLE_DIV = "qc-sample-div"

    # Buttons
    ADD_QC_SAMPLE_BTN = "add-qc-sample-btn"
    REMOVE_QC_SAMPLE_BTN_TYPE = "remove-qc-sample-btn-type"

    # Inputs
    QC_SAMPLE_TYPE = "qc-sample-type"
    QC_SAMPLE_INDEX = "qc-sample-index"

    QC_SAMPLE_NAME_INPUT_TYPE = "qc-sample-name-input-type"

    QC_SAMPLE_ABBREV_INPUT_TYPE = "qc-sample-abbrev-input-type"

    QC_SAMPLE_REMOVE_BTN_TYPE = "qc-sample-remove-btn-type"

    # start
    START_ROUND_ALERT = "start-round-alert"
    START_ROUND_DIV = "start-round-div"
    START_ROUND_INDEX = "start-round-index"

    START_ROUND_POS_SELECT_TYPE = "start-round-pos-select-type"
    START_ROUND_SELECT_TYPE = "start-round-select"

    ADD_START_ROUND_BTN = "add-start-round-btn"
    REMOVE_START_ROUND_BTN_TYPE = "add-start-round-btn-type"

    # end
    END_ROUND_ALERT = "end-round-alert"
    END_ROUND_DIV = "end-round-div"
    END_ROUND_INDEX = "end-round-index"

    END_ROUND_POS_SELECT_TYPE = "end-round-pos-select-type"
    END_ROUND_SELECT_TYPE = "end-round-select"

    ADD_END_ROUND_BTN = "add-end-round-btn"
    REMOVE_END_ROUND_BTN_TYPE = "add-end-round-btn-type"

    # spec round
    SPEC_ROUND_ALERT = "specific-round-alert"
    SPEC_ROUND_DIV = "specific-round-div"
    SPEC_ROUND_INDEX = "specific-round-index"

    SPEC_ROUND_INPUT_TYPE = "specific-round-input-type"
    SPEC_ROUND_SELECT_TYPE = "specific-round-select"

    ADD_SPEC_ROUND_BTN = "add-specific-round-btn"
    REMOVE_SPEC_ROUND_BTN_TYPE = "add-specific-round-btn-type"

    # alternate
    ALTERNATE_ROUND_ALERT = "alternating-round-alert"
    ALTERNATE_ROUND_DIV = "alternating-round-div"
    ALTERNATE_ROUND_INDEX = "alternating-round-index"

    ALTERNATE_ROUND_INPUT_TYPE = "alternating-round-input-type"
    ALTERNATE_ROUND_SELECT_TYPE = "alternating-round-select"

    ADD_ALTERNATE_ROUND_BTN = "add-alternating-round-btn"
    REMOVE_ALTERNATE_ROUND_BTN_TYPE = "add-alternating-round-btn-type"

    # 
    ADDED_TO_LIB_ALERT = "added-to-lib-alert"


class DashIdPlateLib(Enum):
    PLATE_LIB_DAG = "plate-lib-dag"

    REMOVE_PLATE_BTN = "remove-plate-btn"


class DashIdStudy(Enum):
    SAMPLE_LIST_DAG = "sample-list-dag"
    SAMPLE_LIST_DIV = "sample-list-div"
    SAMPLE_LIST_FILENAME_LABEL = "sample-list-filename-label"
    SAMPLE_LIST_UPLOAD = "sample-list-upload"

    PLATE_LAYOUT_DAG = "plate-layout-dag"

    EXAMPLE_LIST_BTN = "example-list-btn"

    PLATE_SELECT = "plate-select"

    SAMPLES_PER_PLATE_SELECT = "samples-per-plate-select"

    RANDOMIZE_SELECT = "randomize-checkbox"

    GROUP_COL_SELECT = "group-col-select"
    GROUP_COL_SELECT_COLLAPSE = "group-col-select-collapse"
    GROUP_COL_SELECT_LABEL = "group-col-select-label-collapse"

    ALLOW_GROUP_SPLIT_SWITCH = "allow-group-split-switch"
    ALLOW_GROUP_SPLIT_COLLAPSE = "allow-group-split-collapse"

    DISTRIBUTE_SAMPLES_BTN = "distribute-samples-plates"

    INPUT_ALERT = "distribute-input-alert"

    PLATE_LAYOUT_SELECT_DAG = "plate-layout-select-"

    SAMPLE_LIST_TAB = "sample-list-tab"
    PLATE_ASSIGN_TAB = "plate-assign-tab"
    PLATE_LAYOUT_TAB = "plate-layout-tab"
    PLATE_EXPORT_TAB = "plate-export-tab"

    PLATE_LAYOUT_GRAPH = "plate-layout-graph"
    PLATE_LAYOUT_TABLE_DIV = "plate-layout-table-div"

    PLATE_ATTRIBUTE_DISTR_GRAPH = "plate-attribute-distribution-graph"

    STUDY_TABS = "study-tabs"

    # Plate layout label
    FIG_PLATE_LABEL = "pl-figure-name-label"
    TABLE_PLATE_LABEL = "pl-table-name-label"

    # Plate layout fig settings
    PL_FIG_COLOR_SELECT = "pl-fig-color-wells-select"
    PL_FIG_LABEL_SELECT = "pl-fig-label-wells-select"

    # Export
    LIST_EXPORT_FIELDS_SELECT = "list-export-fields-select"
    FIG_EXPORT_FORMAT_SELECT = "fig-export-format-select"
    FIG_EXPORT_COLOR_SELECT = "fig-export-color-select"
    FIG_EXPORT_LABEL_SELECT = "fig-export-label-select"

    DOWNLOAD_LISTS_BTN = "download-lists-button"
    LIST_DOWNLOAD = "list-download"

    DOWNLOAD_FIGS_BTN = "download-figures-button"
    FIGS_DOWNLOAD = "figs-download"




import pytest
from pathlib import Path

from src.plate_planner.plate import Plate
from src.plate_planner.plate import Well
from src.plate_planner.study import Study
from src.plate_planner.plate import QCPlate

data_folder = Path("./data/")
config_file = data_folder / "plate_config.toml"

records_file_xlsx = data_folder / "fake_case_control_Npairs_523_Ngroups_5.xlsx"

records_file_csv = data_folder / "fake_case_control_Npairs_523_Ngroups_5.csv"

#@pytest.fixture

# Well CLASS
# def test_should_create_well():
#     well =  Well("A0", (0,0)) 
    
#     assert well.name == "A0"
#     assert well.coordinate == (0,0)
    

# # Plate CLASS
# def test_should_create_plate_from_tuple():
#     plate_96 = Plate((8,12))
    
#     assert plate_96.plate_id == 1
#     assert len(plate_96.rows) == 8
#     assert len(plate_96.columns) == 12
    
    
# def test_should_create_plate_from_dict_args():
#     plate_dim = {"rows":8, 
#                  "columns":12}
    
#     plate_96 = Plate(plate_dim)
    
#     assert len(plate_96.rows) == 8
#     assert len(plate_96.columns) == 12
   
    
# def test_should_create_plate_id_from_key_word_args():
#     plate_dim = {"rows":8, 
#                  "columns":12}
     
#     plate_96 = Plate(plate_dim, plate_id=5)
#     assert plate_96.plate_id == 5
    
    
# def test_should_create_plate_index_coordinates():
#     plate_96 = Plate((8,12))
    
#     plate_96._coordinates = Plate.create_index_coordinates(plate_96.rows, plate_96.columns)
    
#     assert plate_96._coordinates[0] == (7,0)
#     assert plate_96._coordinates[11] == (7,11)
#     assert plate_96._coordinates[84] == (0,0)
#     assert plate_96._coordinates[95] == (0,11)
   
    
# def test_should_create_plate_alphanumerical_coordinates():
#     plate_96 = Plate((8,12))
    
#     plate_96._alphanumerical_coordinates = Plate.create_alphanumerical_coordinates(plate_96.rows, plate_96.columns)[1]
    
#     assert plate_96._alphanumerical_coordinates[0] == "A_1"
#     assert plate_96._alphanumerical_coordinates[11] == "A_12"
#     assert plate_96._alphanumerical_coordinates[84] == "H_1"
#     assert plate_96._alphanumerical_coordinates[95] == "H_12"
    

# @pytest.mark.dependancy(depends=[test_should_create_well,
#                                  test_should_create_plate_index_coordinates, 
#                                  test_should_create_plate_alphanumerical_coordinates])
# def test_should_create_well_objects_in_plate():
#     plate_96 = Plate((8,12))
    
#     plate_96._coordinates = Plate.create_index_coordinates(plate_96.rows, plate_96.columns)
#     plate_96._alphanumerical_coordinates =Plate.create_alphanumerical_coordinates(plate_96.rows, plate_96.columns)[1]
#     Plate.define_empty_wells(plate_96)
    
#     well = plate_96.wells[0]
#     assert well.name == "A_1"
#     assert well.metadata["index"] == 0
#     assert well.coordinate == (7,0)
    
#     well = plate_96.wells[95]
#     assert well.name == "H_12"
#     assert well.metadata["index"]== 95
#     assert well.coordinate == (0,11)
    

# def test_should_return_plate_capacity():
#     plate_96 = Plate((8,12))
#     assert plate_96.capacity == 96 
    

# @pytest.mark.dependancy(depends=[test_should_create_well_objects_in_plate])   
# def test_should_return_length_of_plate(): 
#     plate_96 = Plate((8,12))
#     assert len(plate_96) == 96
      
      
# def test_should_get_well_by_index():
#     plate_96 = Plate((8,12))
    
#     well = plate_96.wells[0]
    
#     assert plate_96[0].name == well.name 
    
    
# def test_should_get_well_by_name():
#     plate_96 = Plate((8,12))
    
#     assert plate_96['A_1'].name ==  plate_96.wells[0].name
#     assert plate_96['A_1'].coordinate == plate_96.wells[0].coordinate
    
#     assert plate_96['B_12'].name ==  plate_96.wells[23].name
#     assert plate_96['B_12'].coordinate == plate_96.wells[23].coordinate
    
#     assert plate_96['H_1'].name ==  plate_96.wells[84].name
#     assert plate_96['H_1'].coordinate == plate_96.wells[84].coordinate
    
#     assert plate_96['H_12'].name ==  plate_96.wells[95].name
#     assert plate_96['H_12'].coordinate == plate_96.wells[95].coordinate


# def test_should_get_well_by_coordinate():
#     plate_96 = Plate((8,12))
    
#     assert plate_96[(7,0)].name == plate_96[0].name
#     assert plate_96[(7,0)].coordinate == plate_96[0].coordinate
    
#     assert plate_96[(6,11)].name == plate_96[23].name
#     assert plate_96[(6,11)].coordinate == plate_96[23].coordinate
    
#     assert plate_96[(0,0)].name == plate_96[84].name
#     assert plate_96[(0,0)].coordinate == plate_96[84].coordinate
    
#     assert plate_96[(0,11)].name == plate_96[95].name
#     assert plate_96[(0,11)].coordinate == plate_96[95].coordinate


# def test_should_get_well_by_key():
#     plate_96 = Plate((8,12))
    
#     assert plate_96[0].name == plate_96[0].name
#     assert plate_96[0].coordinate == plate_96[0].coordinate
    
#     assert plate_96[23].name == plate_96[23].name
#     assert plate_96[23].coordinate == plate_96[23].coordinate
    
#     assert plate_96[84].name == plate_96[84].name
#     assert plate_96[84].coordinate == plate_96[84].coordinate
    
#     assert plate_96[95].name == plate_96[95].name
#     assert plate_96[95].coordinate == plate_96[95].coordinate


# def test_should_set_new_well_by_name():
#     plate_96 = Plate((8,12))
#     test_well = Well("A_12", (7,11))
#     test_well.metadata["barcode"] = 1666
    
#     assert plate_96[11].metadata.get("barcode",0) != 1666
    
#     plate_96["A_12"] = test_well
    
#     assert plate_96[11].name == "A_12"
#     assert plate_96[11].metadata["barcode"] == 1666


# def test_should_set_new_well_by_coordinate():
#     plate_96 = Plate((8,12))
#     test_well = Well("A_12", (7,11))
#     test_well.metadata["barcode"] = 1666
    
#     assert plate_96[11].metadata.get("barcode",0) != 1666
    
#     plate_96[(7,11)] = test_well
    
#     assert plate_96[11].name == "A_12"
#     assert plate_96[11].metadata["barcode"] == 1666


# def test_should_set_new_well_by_index_and_update_well_position():
#     plate_96 = Plate((8,12))
#     test_well = Well("G_1", (15,15))
    
#     plate_96[11] = test_well
    
#     assert plate_96[11].name == "A_12"
#     assert plate_96[(7,11)].coordinate == (7,11)
    
    
# def test_should_get_well_names():
#     plate_96 = Plate((8,12))
    
#     well_names = plate_96.get_metadata("names")
    
#     assert well_names[0] == "A_1"
#     assert well_names[11] == "A_12"
#     assert well_names[84] == "H_1"
#     assert well_names[95] == "H_12"
    
    
# def test_should_get_well_coordinates():
#     plate_96 = Plate((8,12))
    
#     well_crd = plate_96.get_metadata("coordinates")
    
#     assert well_crd[0] == (7,0)
#     assert well_crd[11] == (7,11)
#     assert well_crd[84] == (0,0)
#     assert well_crd[95] == (0,11)
    
# def test_should_get_well_metadata():
#     assert 1 == 2, "Not implemented"
    
    
# def test_should_show_nan_in_get_for_missing_well_keys():
#     assert 1 == 2, "Not implemented"


# # QCplate SUBCLASS
# def test_should_raise_error_on_missing_config_file():
        
#     with pytest.raises(FileExistsError) as exception_info:
#         print(exception_info)
#         qcplate = QCPlate("my_config", (8,12), plate_id=3)
    
#     assert exception_info.value.args[0] == "my_config"

    
# def test_should_create_QCplate():
#     qcplate = QCPlate(config_path, (8,12), plate_id = 3)
    
#     assert qcplate.config["QC"]["run_QC_after_n_specimens"] == 11
#     assert qcplate.plate_id == 3
#     assert qcplate._QC_rounds[1]["sequence"][0] == ("PB",1)
#     assert qcplate._QC_rounds[1]["sequence"][1] == ("EC",2)


# def test_should_get_well_QC_info_from_well_name():
#     qcplate = QCPlate(config_path, (8,12), plate_id = 3)

#     assert qcplate["A_1"].metadata["QC"] == True
    
    
# def test_should_get_well_QC_info_from_well_coord():
#     qcplate = QCPlate(config_path, (8,12), plate_id = 3)

#     assert qcplate[(0,7)].metadata["QC"] == True


# # study CLASS
# @pytest.fixture
# def an_empty_study():
#     study_name = "CNS_tumor_study"
#     return Study(study_name)
    
# @pytest.fixture
# def my_study_with_records(an_empty_study):
#     an_empty_study.load_specimen_records(records_path_xlsx)
#     return an_empty_study
    
# @pytest.fixture
# def my_study(my_study_with_records):
#     my_study_with_records.create_batches(QCPlate(config_path, (8,12)))
#     return my_study_with_records
    
# def test_should_create_empty_study(an_empty_study):
#     study_name = "CNS_tumor_study"
#     my_study = Study(study_name)
    
#     assert my_study.name == study_name
    
    
# def test_should_import_study_records_csv():
#     study_name = "CNS_tumor_study"
#     my_study = Study(study_name)
    
#     my_study.load_specimen_records(records_path_csv)

#     assert my_study.specimen_records_df.empty == False
    
    
# def test_should_import_study_records_xlsx():
#     study_name = "CNS_tumor_study"
#     my_study = Study(study_name)
    
#     my_study.load_specimen_records(records_path_xlsx)

#     assert my_study.specimen_records_df.empty == False   
    
      
# def test_should_fail_import_study_records():
#     study_name = "CNS_tumor_study"
#     my_study = Study(study_name)
    
#     with pytest.raises(FileExistsError) as exception_info:
#         my_study.load_specimen_records("missing_file")
        
#     assert exception_info.value.args[0] == "missing_file"   
    
    
# def test_should_create_batches(my_study):    
#     assert len(my_study.batches) == 14
   
    
# def test_should_get_plate(my_study):
    
#     assert my_study[0].plate_id == 1
#     assert my_study[4].plate_id == 5
#     assert my_study[13].plate_id == 14
    
    
# def test_should_assign_colors_to_wells_by_metadata(my_study):
    
#     plate = my_study[0]
#     rgbs = plate.define_metadata_colors(metadata_key="organ", colormap=plate._colormap)
    
#     plate.assign_well_color(metadata_key = "organ", colormap = plate._colormap)
    
#     assert plate[0].metadata["color"] == rgbs["NaN"]
#     assert plate[2].metadata["color"] == rgbs["Parotid glands"]
#     assert plate[95].metadata["color"] == rgbs["Tendons"]


# def test_should_write_plate_layout_to_txt_file(my_study):
#     plate = my_study[0]
#     plate.to_file()
#     file_path = "Plate_1.txt"
    
#     assert os.path.exists(file_path) == True
    
    
# def test_should_write_plate_layout_to_format_from_file_extension(my_study):
#     plate = my_study[0] 
#     plate.to_file("Plate_1.csv")
#     file_path = "Plate_1.csv"
    
#     assert os.path.exists(file_path) == True
    

# def test_should_write_plate_layout_to_file_given_metadata(my_study): 
#     plate = my_study[0]
#     metadata = ["QC", "pair_ID", "organ", "barcode"]
#     file_path = "Plate_1.csv"
#     plate.to_file(file_path=file_path, metadata_keys=metadata)
   
#     assert os.path.exists(file_path) == True
    
#     #os.remove(file_path)


from dash import html, dcc
import dash_bootstrap_components as dbc

import dash_ag_grid as dag

import plotly.graph_objects as go

# component IDs
from app.callbacks.designer_callback import DashIdPlateDesigner

# views
from app.views.library_view import lib_layout

COLOR_OPTIONS = ["name", "sample_code", "sample_name"]
COLOR_OPTION_DEFAULT = "sample_code"
LABEL_OPTIONS = ["name", "sample_code", "sample_name"]
LABEL_OPTION_DEFAULT = "name"

def create_valid_alert(id):
    return dbc.Alert(
            "Define QC samples above and make sure input values are valid.",
            id=id,
            color="primary",
            dismissable=True,
            is_open=False,
            duration=4000
    )


tab_spec_rounds = html.Div(
    dbc.Row(
        [
            dbc.Col(
                [
                    dbc.Button(
                        "Add",
                        id=DashIdPlateDesigner.ADD_SPEC_ROUND_BTN.value,
                        color="success",
                        outline=True,
                        disabled=False,
                    ),
                ],
                width=2
            ),
            dbc.Col(
                [
                    html.Div(create_valid_alert(DashIdPlateDesigner.SPEC_ROUND_ALERT.value)),
                    html.Div(
                        id=DashIdPlateDesigner.SPEC_ROUND_DIV.value,
                        children=[]
                    )
                ]
            )
        ],
        className="mt-3"
    )
)

tab_start_rounds = html.Div(
    dbc.Row(
        [
            dbc.Col(
                [
                    dbc.Button(
                        "Add",
                        id=DashIdPlateDesigner.ADD_START_ROUND_BTN.value,
                        color="success",
                        outline=True,
                    )
                ],
                width=2
            ),
            dbc.Col(
                [
                    html.Div(create_valid_alert(DashIdPlateDesigner.START_ROUND_ALERT.value)),
                    html.Div(
                        id=DashIdPlateDesigner.START_ROUND_DIV.value,
                        children=[]
                    )
                ]
            )
        ],
        className="mt-3"
    )
)

tab_end_rounds = html.Div(
    dbc.Row(
        [
            dbc.Col(
                [
                    dbc.Button(
                        "Add",
                        id=DashIdPlateDesigner.ADD_END_ROUND_BTN.value,
                        color="success",
                        outline=True
                    )
                ],
                width=2
            ),
            dbc.Col(
                [
                    html.Div(create_valid_alert(DashIdPlateDesigner.END_ROUND_ALERT.value)),
                    html.Div(
                        id=DashIdPlateDesigner.END_ROUND_DIV.value,
                        children=[]
                    )
                ]
            )
        ],
        className="mt-3"
    )
)

tab_alternate_rounds = html.Div(
    dbc.Row(
        [
            dbc.Col(
                [
                    dbc.Button(
                        "Add",
                        id=DashIdPlateDesigner.ADD_ALTERNATE_ROUND_BTN.value,
                        color="success",
                        outline=True
                    )
                ],
                width=2
            ),
            dbc.Col(
                [
                    html.Div(create_valid_alert(DashIdPlateDesigner.ALTERNATE_ROUND_ALERT.value)),
                    html.Div(
                        id=DashIdPlateDesigner.ALTERNATE_ROUND_DIV.value,
                        children=[]
                    )
                ]
            )
        ],
        className="mt-3"
    )
)

qc_sample_def = html.Div(
    [
        dbc.Accordion(
            [
                dbc.AccordionItem(
                    [
                        dbc.Row(
                            [
                                dbc.Col(
                                    [
                                        dbc.Button(
                                            "Add",
                                            id=DashIdPlateDesigner.ADD_QC_SAMPLE_BTN.value,
                                            color="success",
                                            outline=True
                                        ),
                                    ],
                                    width="auto"
                                ),
                                dbc.Col(
                                    [
                                        html.Div(
                                            id=DashIdPlateDesigner.QC_SAMPLE_DIV.value,
                                            children=[

                                            ]
                                        )
                                       
                                    ]
                                )
                            ]
                        )
                    ],
                    title="QC samples"
                ),
                dbc.AccordionItem(
                    [
                        dbc.Row(
                            [
                                dbc.Label("Number of samples between QC rounds", width="auto"),
                                dbc.Col(
                                    [
                                        dbc.Input(id=DashIdPlateDesigner.QC_SAMPLE_SPACING_INPUT.value, type="number", min=0, value=10),
                                    ],
                                    width=2
                                ),
                            ],
                            className="mt-2"
                        ),
                        dbc.Row(
                            [
                                dbc.Col(
                                    [
                                        dbc.Checkbox(
                                            id=DashIdPlateDesigner.START_WITH_QC_CHECKBOX.value,
                                            label="Start with QC round",
                                            value=True
                                        )
                                    ],
                                    width="auto"
                                )
                            ],
                            className="mt-2"
                        ),
                        dbc.Row(
                            [
                                dbc.Col(
                                    [
                                         dbc.Tabs(
                                            [
                                                dbc.Tab(label="Start", tab_id="start_pattern", children=tab_start_rounds),
                                                dbc.Tab(label="End", tab_id="end_pattern", children=tab_end_rounds),
                                                dbc.Tab(label="Specific Round", tab_id="specific_round_patterns", children=tab_spec_rounds),
                                                dbc.Tab(label="Repeat / Alternate", tab_id="alternate_patterns", children=tab_alternate_rounds),
                                                # dbc.Tab(label="Repeat", tab_id="repeat_pattern"),
                                            ],
                                            id="pattern-tabs",
                                            active_tab="start_pattern",
                                        )
                                    ]
                                )
                            ],
                            className="mt-3"
                        )##
                    ],
                    title="Pattern"
                )
            ],
        )
    ]
    )

design_tab = html.Div(
    [
            dbc.Row(
                [
                    dbc.Label("Name", width=2),
                    dbc.Col(
                        [
                            dbc.Input(
                                id=DashIdPlateDesigner.NAME_INPUT.value,
                                value="StudyX"
                            )
                        ],
                        # sm=12, md="auto"
                    ),
                ],
                className="mt-2"
            ),
            dbc.Row(
                [
                    dbc.Label("Size", width=2),
                    dbc.Col(
                        [
                            dbc.Select(
                                id=DashIdPlateDesigner.SIZE_SELECT.value,
                                options=["6", "12", "24", "48", "96", "custom"],
                                value="96"
                            )
                        ],
                        md=12, xl=3
                    ),
                    dbc.Col(
                        [
                            dbc.Collapse(
                                [
                                    dbc.InputGroup(
                                        [
                                            dbc.Input(id=DashIdPlateDesigner.N_ROWS_SELECT.value,type="number", min=1, value=8),
                                            dbc.InputGroupText("rows"),
                                            dbc.Input(id=DashIdPlateDesigner.N_COLUMNS_SELECT.value,type="number", min=1, value=12),
                                            dbc.InputGroupText("columns"),
                                        ]
                                    )
                                ],
                                is_open=True
                            )
                        ],
                        md=12, xl=7
                    ),
                ],
                className="mt-2"
            ),
            dbc.Row(dbc.Col(html.Hr())),
            dbc.Row(dbc.Col(), className="mt-2 mb-2"),
            dbc.Row(dbc.Col(qc_sample_def)),
            dbc.Row(
                [
                    dbc.Col(
                        [
                            dbc.ButtonGroup(
                                [
                                    dbc.Button("Create", id=DashIdPlateDesigner.CREATE_BTN.value, color="success"),
                                    dbc.Button("Preview", id=DashIdPlateDesigner.PREVIEW_BTN.value, color="secondary", outline=True),
                                    dbc.Button("Reset", id=DashIdPlateDesigner.RESET_BTN.value, color="secondary", outline=True),
                                ]
                            )
                        ],
                        width=5
                    ),
                    dbc.Col(
                        [
                            html.Div(
                                dbc.Alert(
                                    id=DashIdPlateDesigner.ADDED_TO_LIB_ALERT.value,
                                    children="kalle",
                                    dismissable=True,
                                    duration=4000,
                                    is_open=False
                                )
                            )
                        ],
                        width=7
                    )
                ],
                className="mt-5"
            ),
    ]
)

library_tab = html.Div(lib_layout)


### Plate preview
plate_design_fig_tab = html.Div(
    [
        dbc.Spinner(
            children=html.Div(
                dcc.Graph(
                    id=DashIdPlateDesigner.PREVIEW_GRAPH.value,
                    figure=go.Figure(),
                    responsive=True,
                    style={"width":"750px", "height": "500px"}
                )
            ),
            type="grow",
            color="secondary"
        ),
        dbc.Row(
            [
                dbc.Label("Color", width="auto"),
                dbc.Col(
                    [
                        dbc.Select(
                            id=DashIdPlateDesigner.COLOR_SELECT.value,
                            options=COLOR_OPTIONS,
                            value=COLOR_OPTION_DEFAULT
                        )
                    ],                                    
                ),
                dbc.Label("Label", width="auto"),
                dbc.Col(
                    [
                        dbc.Select(
                            id=DashIdPlateDesigner.LABEL_SELECT.value,
                            options=LABEL_OPTIONS,
                            value=LABEL_OPTION_DEFAULT
                        )
                    ],
                )
            ],
            className="ms-4 me-4",
            style={"width": "700px"}
        )
    ]
)

### table
column_defs = [
    {
        "headerName": "Well Index",
        "field": "index"
    },
    {
        "headerName": "Well",
        "field": "name"
    },
    {
        "headerName": "Sample Type",
        "field": "sample_type"
    },
        {
        "headerName": "Sample Code",
        "field": "sample_code"
    }
]

plate_preview_dag = dag.AgGrid(
    id=DashIdPlateDesigner.PREVIEW_DAG.value,
    rowData=[],
    columnSize="responsiveSizeToFit",

    dashGridOptions={
        "rowSelection": "single",
    },

    columnDefs=column_defs,

    className="dbc ag-theme-alpine ag-theme-alpine2",

    style={"width": "100%", "height": "500px"},
)

plate_design_table_tab = html.Div(
    [
        dbc.Row(
            [
                dbc.Col(
                    [
                        plate_preview_dag
                    ]
                )
            ],
            
        ),
        
    ]
)


### MAIN
designer_layout = html.Div(
    [
        dbc.Row(
            [
                dbc.Col(
                    [
                        html.H4("Plate templates"),
                    ]
                )
            ],
            className="mt-2 ms-4 me-4"
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dbc.Tabs(
                            [
                                dbc.Tab(label="Design", children=design_tab),
                                dbc.Tab(label="Library", children=library_tab)
                            ]
                        )
                    ],
                    sm=12, md=12, xl=6
                ),
                dbc.Col(
                    [
                        dbc.Tabs(
                            [
                                dbc.Tab(label="Figure", children=plate_design_fig_tab),
                                dbc.Tab(label="Table", children=plate_design_table_tab),
                            ],
                            # className="mt-2"
                        )
                    ],
                    sm=12, md=12, xl=6
                )
            ],
            className="mt-2 ms-4 me-4"
        ),
    ]
)